package build

import (
	"sync"

	"git.akyoto.dev/cli/q/src/build/asm"
	fail "git.akyoto.dev/cli/q/src/errors"
)

// compile waits for the scan to finish and compiles all functions.
func compile(functions <-chan *Function, errors <-chan error) (Result, error) {
	result := Result{
		Unused: map[string]*Function{},
	}

	for functions != nil || errors != nil {
		select {
		case err, ok := <-errors:
			if !ok {
				errors = nil
				continue
			}

			return result, err

		case function, ok := <-functions:
			if !ok {
				functions = nil
				continue
			}

			result.Unused[function.Name] = function
		}
	}

	compileFunctions(result.Unused)

	for _, function := range result.Unused {
		if function.Error != nil {
			return result, function.Error
		}

		result.instructionCount += len(function.Assembler.Instructions)
	}

	main, exists := result.Unused["main"]

	if !exists {
		return result, fail.MissingMainFunction
	}

	result.Used = append(result.Used, main)
	delete(result.Unused, "main")
	result.findCalls(main)

	return result, nil
}

func (result *Result) findCalls(f *Function) {
	for _, x := range f.Assembler.Instructions {
		if x.Mnemonic != asm.CALL {
			continue
		}

		name := x.Data.(*asm.Label).Name
		called, exists := result.Unused[name]

		if !exists {
			continue
		}

		result.Used = append(result.Used, called)
		delete(result.Unused, name)
		result.findCalls(called)
	}
}

// compileFunctions starts a goroutine for each function compilation and waits for completion.
func compileFunctions(functions map[string]*Function) {
	wg := sync.WaitGroup{}

	for _, function := range functions {
		wg.Add(1)

		go func() {
			defer wg.Done()
			function.Compile()
		}()
	}

	wg.Wait()
}
